from typing import Dict, List, Optional, Union

from unify import BASE_URL
from unify.utils import _requests

from ...utils.helpers import _check_response, _validate_api_key

# Projects #
# ---------#


def create_project(
    name: str,
    *,
    overwrite: Union[bool, str] = False,
    api_key: Optional[str] = None,
    is_versioned: bool = True,
) -> Dict[str, str]:
    """
    Creates a logging project and adds this to your account. This project will have
    a set of logs associated with it.

    Args:
        name: A unique, user-defined name used when referencing the project.

        overwrite: Controls how to handle existing projects with the same name.
            If False (default), raises an error if project exists.
            If True, deletes the entire existing project before creating new one.
            If "logs", only deletes the project's logs before creating.
            If "contexts", only deletes the project's contexts before creating.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

        is_versioned: Whether the project is tracked via version control.

    Returns:
        A message indicating whether the project was created successfully.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    body = {"name": name, "is_versioned": is_versioned}
    if overwrite:
        if name in list_projects(api_key=api_key):
            if overwrite == "logs":
                return delete_project_logs(name=name, api_key=api_key)
            elif overwrite == "contexts":
                return delete_project_contexts(name=name, api_key=api_key)
            else:
                delete_project(name=name, api_key=api_key)
    response = _requests.post(BASE_URL + "/project", headers=headers, json=body)
    _check_response(response)
    return response.json()


def rename_project(
    name: str,
    new_name: str,
    *,
    api_key: Optional[str] = None,
) -> Dict[str, str]:
    """
    Renames a project from `name` to `new_name` in your account.

    Args:
        name: Name of the project to rename.

        new_name: A unique, user-defined name used when referencing the project.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A message indicating whether the project was successfully renamed.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    body = {"name": new_name}
    response = _requests.patch(
        BASE_URL + f"/project/{name}",
        headers=headers,
        json=body,
    )
    _check_response(response)
    return response.json()


def delete_project(
    name: str,
    *,
    api_key: Optional[str] = None,
) -> str:
    """
    Deletes a project from your account.

    Args:
        name: Name of the project to delete.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        Whether the project was successfully deleted.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    response = _requests.delete(BASE_URL + f"/project/{name}", headers=headers)
    _check_response(response)
    return response.json()


def delete_project_logs(
    name: str,
    *,
    api_key: Optional[str] = None,
) -> None:
    """
    Deletes all logs from a project.

    Args:
        name: Name of the project to delete logs from.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    response = _requests.delete(BASE_URL + f"/project/{name}/logs", headers=headers)
    _check_response(response)
    return response.json()


def delete_project_contexts(
    name: str,
    *,
    api_key: Optional[str] = None,
) -> None:
    """
    Deletes all contexts and their associated logs from a project

    Args:
        name: Name of the project to delete contexts from.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    response = _requests.delete(BASE_URL + f"/project/{name}/contexts", headers=headers)
    _check_response(response)
    return response.json()


def list_projects(
    *,
    api_key: Optional[str] = None,
) -> List[str]:
    """
    Returns the names of all projects stored in your account.

    Args:
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        List of all project names.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    response = _requests.get(BASE_URL + "/projects", headers=headers)
    _check_response(response)
    return response.json()


def commit_project(
    name: str,
    commit_message: str,
    *,
    api_key: Optional[str] = None,
) -> Dict[str, str]:
    """
    Creates a commit for the entire project, saving a snapshot of all versioned contexts.

    Args:
        name: Name of the project to commit.
        commit_message: A description of the changes being saved.
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the new commit_hash.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    body = {"commit_message": commit_message}
    response = _requests.post(
        BASE_URL + f"/project/{name}/commit",
        headers=headers,
        json=body,
    )
    _check_response(response)
    return response.json()


def rollback_project(
    name: str,
    commit_hash: str,
    *,
    api_key: Optional[str] = None,
) -> Dict[str, str]:
    """
    Rolls back the entire project to a specific commit.

    Args:
        name: Name of the project to roll back.
        commit_hash: The hash of the commit to restore.
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A message indicating the success of the rollback operation.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    body = {"commit_hash": commit_hash}
    response = _requests.post(
        BASE_URL + f"/project/{name}/rollback",
        headers=headers,
        json=body,
    )
    _check_response(response)
    return response.json()


def get_project_commits(name: str, *, api_key: Optional[str] = None) -> List[Dict]:
    """
    Retrieves the commit history for a project.

    Args:
        name: Name of the project.
        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A list of dictionaries, each representing a commit.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    response = _requests.get(BASE_URL + f"/project/{name}/commits", headers=headers)
    _check_response(response)
    return response.json()



import functools
from typing import List, Optional

from unify import BASE_URL
from unify.utils import _requests

from ...utils.helpers import _res_to_list, _validate_api_key


@functools.lru_cache(maxsize=32)
def list_providers(
    model: Optional[str] = None,
    *,
    api_key: Optional[str] = None,
) -> List[str]:
    """
    Get a list of available providers, either in total or for a specific model.

    Args:
        model: If specified, returns the list of providers supporting this model.
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A list of provider names associated with the model if successful, otherwise an
        empty list.
    Raises:
        BadRequestError: If there was an HTTP error.
        ValueError: If there was an error parsing the JSON response.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/providers"
    if model:
        kw = dict(headers=headers, params={"model": model})
    else:
        kw = dict(headers=headers)
    response = _requests.get(url, **kw)
    if response.status_code != 200:
        raise Exception(response.json())
    return _res_to_list(response)


@functools.lru_cache(maxsize=32)
def list_models(
    provider: Optional[str] = None,
    *,
    api_key: Optional[str] = None,
) -> List[str]:
    """
    Get a list of available models, either in total or for a specific provider.

    Args:
        provider: If specified, returns the list of models supporting this provider.
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A list of available model names if successful, otherwise an empty list.
    Raises:
        BadRequestError: If there was an HTTP error.
        ValueError: If there was an error parsing the JSON response.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/models"
    if provider:
        kw = dict(headers=headers, params={"provider": provider})
    else:
        kw = dict(headers=headers)
    response = _requests.get(url, **kw)
    if response.status_code != 200:
        raise Exception(response.json())
    return _res_to_list(response)


@functools.lru_cache(maxsize=32)
def list_endpoints(
    model: Optional[str] = None,
    provider: Optional[str] = None,
    *,
    api_key: Optional[str] = None,
) -> List[str]:
    """
    Get a list of available endpoint, either in total or for a specific model or
    provider.

    Args:
        model: If specified, returns the list of endpoint supporting this model.
        provider: If specified, returns the list of endpoint supporting this provider.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A list of endpoint names if successful, otherwise an empty list.
    Raises:
        BadRequestError: If there was an HTTP error.
        ValueError: If there was an error parsing the JSON response.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/endpoints"
    if model and provider:
        raise ValueError("Please specify either model OR provider, not both.")
    elif model:
        kw = dict(headers=headers, params={"model": model})
        return _res_to_list(
            _requests.get(url, headers=headers, params={"model": model}),
        )
    elif provider:
        kw = dict(headers=headers, params={"provider": provider})
    else:
        kw = dict(headers=headers)
    response = _requests.get(url, **kw)
    if response.status_code != 200:
        raise Exception(response.json())
    return _res_to_list(response)



import importlib
import sys
from typing import Callable, List

import unify


class TraceLoader(importlib.abc.Loader):
    def __init__(self, original_loader, filter: Callable = None):
        self._original_loader = original_loader
        self.filter = filter

    def create_module(self, spec):
        return self._original_loader.create_module(spec)

    def exec_module(self, module):
        self._original_loader.exec_module(module)
        unify.traced(module, filter=self.filter)


class TraceFinder(importlib.abc.MetaPathFinder):
    def __init__(self, targets: List[str], filter: Callable = None):
        self.targets = targets
        self.filter = filter

    def find_spec(self, fullname, path, target=None):
        for target_module in self.targets:
            if not fullname.startswith(target_module):
                return None

        original_sys_meta_path = sys.meta_path[:]
        sys.meta_path = [
            finder for finder in sys.meta_path if not isinstance(finder, TraceFinder)
        ]
        try:
            spec = importlib.util.find_spec(fullname, path)
            if spec is None:
                return None
        finally:
            sys.meta_path = original_sys_meta_path

        if spec.origin is None or not spec.origin.endswith(".py"):
            return None

        spec.loader = TraceLoader(spec.loader, filter=self.filter)
        return spec


def install_tracing_hook(targets: List[str], filter: Callable = None):
    """Install an import hook that wraps imported modules with the traced decorator.

    This function adds a TraceFinder to sys.meta_path that will intercept module imports
    and wrap them with the traced decorator. The hook will only be installed if one
    doesn't already exist.

    Args:
        targets: List of module name prefixes to target for tracing. Only modules
            whose names start with these prefixes will be wrapped.

        filter: A filter function that is passed to the traced decorator.

    """
    if not any(isinstance(finder, TraceFinder) for finder in sys.meta_path):
        sys.meta_path.insert(0, TraceFinder(targets, filter))


def disable_tracing_hook():
    """Remove the tracing import hook from sys.meta_path.

    This function removes any TraceFinder instances from sys.meta_path, effectively
    disabling the tracing functionality for subsequent module imports.

    """
    for finder in sys.meta_path:
        if isinstance(finder, TraceFinder):
            sys.meta_path.remove(finder)



from __future__ import annotations

import json

from ...utils.helpers import _validate_api_key
from .logs import *

# Parameters #
# -----------#


def get_param_by_version(
    field: str,
    version: Union[str, int],
    api_key: Optional[str] = None,
) -> Any:
    """
    Gets the parameter by version.

    Args:
        field: The field of the parameter to get.

        version: The version of the parameter to get.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        The parameter by version.
    """
    api_key = _validate_api_key(api_key)
    version = str(version)
    filter_exp = f"version({field}) == {version}"
    return get_logs(filter=filter_exp, limit=1, api_key=api_key)[0].params[field][1]


def get_param_by_value(
    field: str,
    value: Any,
    api_key: Optional[str] = None,
) -> Any:
    """
    Gets the parameter by value.

    Args:
        field: The field of the parameter to get.

        value: The value of the parameter to get.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        The parameter by version.
    """
    api_key = _validate_api_key(api_key)
    filter_exp = f"{field} == {json.dumps(value)}"
    return get_logs(filter=filter_exp, limit=1, api_key=api_key)[0].params[field][0]


def get_source() -> str:
    """
    Extracts the source code for the file from where this function was called.

    Returns:
        The source code for the file, as a string.
    """
    frame = inspect.getouterframes(inspect.currentframe())[1]
    with open(frame.filename, "r") as file:
        source = file.read()
    return f"```python\n{source}\n```"


# Experiments #
# ------------#


def get_experiment_name(version: int, api_key: Optional[str] = None) -> str:
    """
    Gets the experiment name (by version).

    Args:
        version: The version of the experiment to get.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        The experiment name with said version.
    """
    experiments = get_groups(key="experiment", api_key=api_key)
    if not experiments:
        return None
    elif version < 0:
        version = len(experiments) + version
    if str(version) not in experiments:
        return None
    return experiments[str(version)]


def get_experiment_version(name: str, api_key: Optional[str] = None) -> int:
    """
    Gets the experiment version (by name).

    Args:
        name: The name of the experiment to get.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        The experiment version with said name.
    """
    experiments = get_groups(key="experiment", api_key=api_key)
    if not experiments:
        return None
    experiments = {v: k for k, v in experiments.items()}
    if name not in experiments:
        return None
    return int(experiments[name])



from typing import List, Type, Union

from openai.types.chat import ChatCompletion
from unify.universal_api.types import Prompt

# Upcasting


def _usr_msg_to_prompt(user_message: str) -> Prompt:
    return Prompt(user_message)


def _bool_to_float(boolean: bool) -> float:
    return float(boolean)


# Downcasting


def _prompt_to_usr_msg(prompt: Prompt) -> str:
    return prompt.messages[-1]["content"]


def _chat_completion_to_assis_msg(chat_completion: ChatCompletion) -> str:
    return chat_completion.choices[0].message.content


def _float_to_bool(float_in: float) -> bool:
    return bool(float_in)


# Cast Dict

_CAST_DICT = {
    str: {Prompt: _usr_msg_to_prompt},
    Prompt: {
        str: _prompt_to_usr_msg,
    },
    ChatCompletion: {str: _chat_completion_to_assis_msg},
    bool: {
        float: _bool_to_float,
    },
    float: {
        bool: _float_to_bool,
    },
}


def _cast_from_selection(
    inp: Union[str, bool, float, Prompt, ChatCompletion],
    targets: List[Union[float, Prompt, ChatCompletion]],
) -> Union[str, bool, float, Prompt, ChatCompletion]:
    """
    Upcasts the input if possible, based on the permitted upcasting targets provided.

    Args:
        inp: The input to cast.

        targets: The set of permitted upcasting targets.

    Returns:
        The input after casting to the new type, if it was possible.
    """
    input_type = type(inp)
    assert input_type in _CAST_DICT, (
        "Cannot upcast input {} of type {}, because this type is not in the "
        "_CAST_DICT, meaning there are no functions for casting this type."
    )
    cast_fns = _CAST_DICT[input_type]
    targets = [target for target in targets if target in cast_fns]
    assert len(targets) == 1, "There must be exactly one valid casting target."
    to_type = targets[0]
    return cast_fns[to_type](inp)


# Public function


def cast(
    inp: Union[str, bool, float, Prompt, ChatCompletion],
    to_type: Union[
        Type[Union[str, bool, float, Prompt, ChatCompletion]],
        List[Type[Union[str, bool, float, Prompt, ChatCompletion]]],
    ],
) -> Union[str, bool, float, Prompt, ChatCompletion]:
    """
    Cast the input to the specified type.

    Args:
        inp: The input to cast.

        to_type: The type to cast the input to.

    Returns:
        The input after casting to the new type.
    """
    if isinstance(to_type, list):
        return _cast_from_selection(inp, to_type)
    input_type = type(inp)
    if input_type is to_type:
        return inp
    return _CAST_DICT[input_type][to_type](inp)


def try_cast(
    inp: Union[str, bool, float, Prompt, ChatCompletion],
    to_type: Union[
        Type[Union[str, bool, float, Prompt, ChatCompletion]],
        List[Type[Union[str, bool, float, Prompt, ChatCompletion]]],
    ],
) -> Union[str, bool, float, Prompt, ChatCompletion]:
    # noinspection PyBroadException
    try:
        return cast(inp, to_type)
    except:
        return inp



import asyncio
import atexit
import logging
import os
import threading
from concurrent.futures import TimeoutError
from typing import List

import aiohttp
from unify import BASE_URL

# Configure logging based on environment variable
ASYNC_LOGGER_DEBUG = os.getenv("UNIFY_ASYNC_LOGGER_DEBUG", "false").lower() in (
    "true",
    "1",
)


class AsyncLoggerManager:
    def __init__(
        self,
        *,
        name: str = "unknown",
        base_url: str = BASE_URL,
        api_key: str = os.getenv("UNIFY_KEY"),
        num_consumers: int = 256,
        max_queue_size: int = 10000,
    ):
        self.name = f"UnifyAsyncLogger.{name}"
        self.loop = asyncio.new_event_loop()
        self.queue = None
        self.consumers: List[asyncio.Task] = []
        self.num_consumers = num_consumers
        self.start_flag = threading.Event()
        self.shutting_down = False
        self.thread_raised_exception = threading.Event()
        self.max_queue_size = max_queue_size
        self.logger = logging.getLogger(self.name)
        if ASYNC_LOGGER_DEBUG:
            self.logger.setLevel(logging.DEBUG)

        # Register shutdown handler
        atexit.register(self.stop_sync, immediate=False)

        headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
            "accept": "application/json",
        }
        url = base_url + "/"
        connector = aiohttp.TCPConnector(limit=num_consumers // 2, loop=self.loop)
        self.session = aiohttp.ClientSession(
            url,
            headers=headers,
            loop=self.loop,
            connector=connector,
        )

        self.thread = threading.Thread(
            target=self._run_loop,
            daemon=True,
            name=self.name,
        )
        self.thread.start()
        self.start_flag.wait()
        if self.thread_raised_exception.is_set():
            raise RuntimeError(f"{self.name} thread failed to start")
        self.callbacks = []

    def register_callback(self, fn):
        self.callbacks.append(fn)

    def clear_callbacks(self):
        self.callbacks = []

    def _notify_callbacks(self):
        for fn in self.callbacks:
            fn()

    async def _join(self):
        if self.queue is None:
            return
        await self.queue.join()

    def join(self):
        try:
            future = asyncio.run_coroutine_threadsafe(self._join(), self.loop)
            while True:
                try:
                    future.result(timeout=0.5)
                    break
                except (asyncio.TimeoutError, TimeoutError):
                    self.logger.debug(
                        f"Join waiting for {self.queue._unfinished_tasks} tasks to complete",
                    )
                    continue
        except Exception as e:
            self.logger.error(f"Error in join: {e}")
            raise e

    async def _main_loop(self):
        self.start_flag.set()
        self.logger.debug(f"Spawning {self.num_consumers} consumers")
        await asyncio.gather(*self.consumers, return_exceptions=True)

    def _run_loop(self):
        asyncio.set_event_loop(self.loop)
        self.queue = asyncio.Queue(maxsize=self.max_queue_size)

        for _ in range(self.num_consumers):
            self.consumers.append(self._log_consumer())

        try:
            self.loop.run_until_complete(self._main_loop())
        except Exception as e:
            self.shutting_down = True
            self.thread_raised_exception.set()
            self.start_flag.set()
            raise e
        finally:
            self.loop.close()

    async def _consume_create(self, body, future, idx):
        async with self.session.post("logs", json=body) as res:
            if res.status != 200:
                txt = await res.text()
                self.logger.error(
                    f"Failed to create log {idx} {res.status}: {txt}",
                )
                return
            res_json = await res.json()
            self.logger.debug(
                f"Created log {res_json['log_event_ids'][0]} with status {res.status}",
            )
            future.set_result(res_json["log_event_ids"][0])

    async def _consume_update(self, body, future, idx):
        if not future.done():
            await future
        body["logs"] = [future.result()]
        async with self.session.put("logs", json=body) as res:
            if res.status != 200:
                txt = await res.text()
                self.logger.error(
                    f"Failed to update log {idx} {body['logs'][0]} {res.status}: {txt}",
                )
                return
            res_json = await res.json()
            self.logger.debug(
                f"Updated log {res_json['log_event_ids'][0]} with status {res.status}",
            )

    async def _log_consumer(self):
        while True:
            try:
                event = await self.queue.get()
                idx = self.queue.qsize() + 1
                self.logger.debug(f"'{event['type']}' processing {idx}")
                if event["type"] == "create":
                    await self._consume_create(event["_data"], event["future"], idx)
                elif event["type"] == "update":
                    await self._consume_update(event["_data"], event["future"], idx)
                else:
                    raise Exception(f"Unknown event type: {event['type']}")
            except Exception as e:
                event["future"].set_exception(e)
                self.logger.error(f"Error in consumer: {e}")
                raise e
            finally:
                self.queue.task_done()
                self._notify_callbacks()

    def log_create(
        self,
        project: str,
        context: str,
        params: dict,
        entries: dict,
    ) -> asyncio.Future:
        if self.shutting_down:
            self.logger.warning(f"Not running, skipping log create")
            return None
        fut = self.loop.create_future()
        event = {
            "_data": {
                "project": project,
                "context": context,
                "params": params,
                "entries": entries,
            },
            "type": "create",
            "future": fut,
        }
        asyncio.run_coroutine_threadsafe(self.queue.put(event), self.loop).result()
        return fut

    def log_update(
        self,
        project: str,
        context: str,
        future: asyncio.Future,
        mode: str,
        overwrite: bool,
        data: dict,
    ) -> None:
        if self.shutting_down:
            self.logger.warning(f"Not running, skipping log update")
            return
        event = {
            "_data": {
                mode: data,
                "project": project,
                "context": context,
                "overwrite": overwrite,
            },
            "type": "update",
            "future": future,
        }
        asyncio.run_coroutine_threadsafe(self.queue.put(event), self.loop).result()

    def clear_queue(self):
        if self.queue is None:
            return

        orig_size = self.queue.qsize()
        while not self.queue.empty():
            self.queue.get_nowait()
            self.queue.task_done()
        self.logger.debug(f"{orig_size} log requests cleared")

    def stop_sync(self, immediate=False):
        if self.shutting_down:
            self.logger.debug(f"Already shutting down, skipping stop")
            return

        self.shutting_down = True
        if immediate:
            self.logger.debug(f"Shutting down immediately")
            self.loop.stop()
        else:
            self.logger.debug(f"Shutting down gracefully")
            self.join()



import datetime
from typing import List, Optional

import unify

from ..utils.helpers import _validate_api_key


def with_logging(
    model_fn: Optional[callable] = None,
    *,
    endpoint: str,
    tags: Optional[List[str]] = None,
    timestamp: Optional[datetime.datetime] = None,
    log_query_body: bool = True,
    log_response_body: bool = True,
    api_key: Optional[str] = None,
):
    """
    Wrap a local model callable with logging of the queries.

    Args:
        model_fn: The model callable to wrap logging around.
        endpoint: The endpoint name to give to this local callable.
        tags: Tags for later filtering.
        timestamp: A timestamp (if not set, will be the time of sending).
        log_query_body: Whether or not to log the query body.
        log_response_body: Whether or not to log the response body.
        api_key: If specified, unify API key to be used. Defaults to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A new callable, but with logging added every time the function is called.

    Raises:
        requests.HTTPError: If the API request fails.
    """
    _tags = tags
    _timestamp = timestamp
    _log_query_body = log_query_body
    _log_response_body = log_response_body
    api_key = _validate_api_key(api_key)

    # noinspection PyShadowingNames
    def model_fn_w_logging(
        *args,
        tags: Optional[List[str]] = None,
        timestamp: Optional[datetime.datetime] = None,
        log_query_body: bool = True,
        log_response_body: bool = True,
        **kwargs,
    ):
        if len(args) != 0:
            raise Exception(
                "When logging queries for a local model, all arguments to "
                "the model callable must be provided as keyword arguments. "
                "Positional arguments are not supported. This is so the "
                "query body dict can be fully populated with keys for each "
                "entry.",
            )
        query_body = kwargs
        response = model_fn(**query_body)
        if not isinstance(response, dict):
            response = {"response": response}
        kw = dict(
            endpoint=endpoint,
            query_body=query_body,
            response_body=response,
            tags=tags,
            timestamp=timestamp,
            api_key=api_key,
        )
        if log_query_body:
            if not log_response_body:
                del kw["response_body"]
            unify.log_query(**kw)
        return response

    return model_fn_w_logging
